#ifndef EAVESDROPPER_INFOWINDOW_H
#define EAVESDROPPER_INFOWINDOW_H

// ***************************************************************************
// System Includes
// ***************************************************************************

#include <be/interface/TextView.h>
#include <be/interface/Window.h>
#include <be/media/MediaRoster.h>
#include <be/support/List.h>

// ***************************************************************************
// Project Includes
// ***************************************************************************

#include "StyleGuideDefs.h"         // Standard type and constant definitions.

// ***************************************************************************
// External Name Declarations
// ***************************************************************************

class ConnectorView;
class AbstractNodeView;
class DormantNodeView;
class LiveNodeView;

// ***************************************************************************
// External Variable Declarations
// ***************************************************************************

// ***************************************************************************
// Constant Definitions
// ***************************************************************************

// ***************************************************************************
// Type Definitions
// ***************************************************************************

// ***************************************************************************
// Class Definition
// ***************************************************************************

/**  
 * The InfoWindow class is Eavesdropper's Information window, which shows
 * detailed information about the selected node or connector. It also handles
 * the selection of these items.
 */
class InfoWindow : public BWindow
{
public:
	// Constant Definitions ==================================================

	// Type Definitions ======================================================

	// Attribute Definitions =================================================

	// Member Function Definitions ===========================================
 
	/**
	 * Constructor.
	 *
	 * @param mediaRoster Pointer to the Media Roster.
	 */
	InfoWindow(BMediaRoster* mediaRoster); 

	/** Automatically generated copy constructor. */
	// InfoWindow(const InfoWindow& old);

	/** Automatically generated destructor. */
	// virtual ~InfoWindow(void);

	/** Automatically generated assignment operator. */
	// InfoWindow& operator=(InfoWindow& right);

	/** Responds to a notification that the window has moved. */
	virtual void FrameMoved(BPoint screenPoint);
	
	/** Responds to a notification that the window has been resized. */
	virtual void FrameResized(float width, float height);
	
	/** Selects a node. */
	void SelectNode(AbstractNodeView* node);

	/** Selects a connector. */
	void SelectConnector(ConnectorView* connector);
	
	/** Called when there is nothing to select. */
	void SelectNothing();
	
protected:
	// Constant Definitions ==================================================

	// Type Definitions ======================================================

	// Attribute Definitions =================================================

	// Member Function Definitions ===========================================

private:
	// Constant Definitions ==================================================

	// Type Definitions ======================================================

	// Attribute Definitions =================================================

	/** Pointer to the Media Roster. */
	BMediaRoster* mediaRoster;

	/** The text view that shows the information. */
	BTextView* infoView;

	/** The node that is currently selected. */
	AbstractNodeView* selectedNode;

	/** The connector that is currently selected. */
	ConnectorView* selectedConnector;
	
	// Member Function Definitions ===========================================

	/** Saves the window's frame into the settings file. */	
	void SaveSettings(void);

	/** Shows the information about a certain live node. */
	void ShowLiveNode(LiveNodeView* liveNodeView);

	/** Shows the information about a certain dormant node. */
	void ShowDormantNode(DormantNodeView* dormantNodeView);

	/** Shows the information about a certain connector. */
	void ShowConnector(ConnectorView* connectorView);
	
	/** Writes a node_kind object into the text view. */	
	void WriteNodeKind(uint32 kind);

	/** Writes a media_file_format object into the text view. */
	void WriteMediaFileFormat(media_file_format* fileFormat);
	
	/** Writes a media_source object into the text view. */
	void WriteMediaSource(media_source mediaSource);
	
	/** Writes a media_destination object into the text view. */
	void WriteMediaDestination(media_destination mediaDestination);

	/** Writes part of a media_format object into the text view. */
	void WriteMediaFormatSummary(media_format mediaFormat);

	/** Writes a media_format object into the text view. */
	void WriteMediaFormat(media_format mediaFormat);
	
	/** Writes a media_type object into the text view. */
	void WriteMediaType(media_type mediaType);
	
	/** Writes a type_code object into the text view. */
	void WriteTypeCode(type_code typeCode);
	
	/** Writes a color_space object into the text view. */
	void WriteColorSpace(color_space colorSpace);

	/** Writes a media_raw_audio_format object into the text view. */
	void WriteRawAudioFormat(media_raw_audio_format format);
	
	/** Writes a media_raw_video_format object into the text view. */
	void WriteRawVideoFormat(media_raw_video_format format);
};

// ***************************************************************************
// Global Variable Declarations
// ***************************************************************************

// ***************************************************************************
// Inline Function Definitions
// ***************************************************************************

// ***************************************************************************
// Helper Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Function Declarations
// ***************************************************************************

#endif // EAVESDROPPER_INFOWINDOW_H
