/*!
 * [TODO] Automatically recognize new nodes, new connections, and so on.
 *
 * Implement this using BMediaRoster::StartWatching().
 */

// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

#include <be/app/Application.h>	// be_app

#include <be/interface/Menu.h>
#include <be/interface/MenuBar.h>
#include <be/interface/MenuItem.h>
#include <be/interface/ScrollView.h>

#include <be/media/MediaAddOn.h>
#include <be/media/MediaNode.h>

#include <be/storage/File.h>

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "Connection.h"
#include "ConnectorView.h"
#include "EavesdropperMain.h"
#include "Fail.h"
#include "InfoWindow.h"
#include "MainWindow.h"
#include "LiveNodeView.h"
#include "DormantNodeView.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

MainWindow::MainWindow(BMediaRoster* mediaRoster, InfoWindow* infoWindow) 
	: BWindow(BRect(10, 50, 380, 470), 
		"Eavesdropper", 
		B_DOCUMENT_WINDOW, 
		0)                       // No flags.
{
	// Initialize attributes.
	this->mediaRoster = mediaRoster;
	this->infoWindow = infoWindow;
	liveNodes = 0;
	connections = 0;
	dormantNodes = 0;

	// -----------------------------------------------------------------------

	// Read the window's frame rectangle from the settings file.
	BFile file(SETTINGS_FILE, B_READ_ONLY);
	if (file.InitCheck() == B_OK)
	{
		if (file.Lock() == B_OK)
		{
			BRect rect;
			if (file.ReadAttr("MainWindowFrame",
				B_RECT_TYPE,
				0,
				&rect,
				sizeof(BRect)) > 0)
			{
				MoveTo(rect.left,rect.top);
				ResizeTo(rect.Width(),rect.Height());
			}
      
			file.Unlock();
		}
	}

	// -----------------------------------------------------------------------

	// Build the menu bar.
	BMenuBar* menuBar = new BMenuBar(Bounds(), "MenuBar");

	BMenu* menu = new BMenu("Options");

	menu->AddItem(new BMenuItem("Refresh",
		new BMessage(MSG_REFRESH),
		'R'));

	menu->AddItem(new BMenuItem("Clean up",
		new BMessage(MSG_CLEAN_UP),
		'K'));

	menu->AddSeparatorItem();

	menu->AddItem(new BMenuItem("About" B_UTF8_ELLIPSIS,
		new BMessage(B_ABOUT_REQUESTED)));
		
	menu->AddSeparatorItem();

	menu->AddItem(new BMenuItem("Quit",
		new BMessage(B_QUIT_REQUESTED),
		'Q'));

	menuBar->AddItem(menu);

	AddChild(menuBar);

	// -----------------------------------------------------------------------
	// Build the dormant nodes view.
	
	dormantNodesView = new DormantNodesView(
		BRect(0,
			menuBar->Frame().bottom + 1,
			100 - B_V_SCROLL_BAR_WIDTH - 1,
			Bounds().Height() + 1));
			
    BScrollView* dormantNodesScrollView = new BScrollView(
		"DormantNodesScrollView",
		dormantNodesView,
		B_FOLLOW_TOP_BOTTOM,
		0,
		false,
		true,
		B_NO_BORDER);
		
	AddChild(dormantNodesScrollView);

	// -----------------------------------------------------------------------
	// Build the live nodes view.

	liveNodesView = new LiveNodesView(BRect(101,
			menuBar->Frame().bottom + 1,
			Bounds().Width() - B_V_SCROLL_BAR_WIDTH,
			Bounds().Height() - B_H_SCROLL_BAR_HEIGHT));

    BScrollView* liveNodesScrollView = new BScrollView(
		"LiveNodesScrollView",
		liveNodesView,
		B_FOLLOW_ALL_SIDES,
		0,
		true,
		true,
		B_NO_BORDER);
		
	AddChild(liveNodesScrollView);

	// -----------------------------------------------------------------------

	// Get the views of the live nodes and add them to the live nodes view.
	Refresh();
	CleanUp();
}
 
// ===========================================================================
// Destructor
// ===========================================================================

MainWindow::~MainWindow()
{
	GetRidOfStuff();
}
 
// ===========================================================================
// QuitRequested
// ===========================================================================

bool 
MainWindow::QuitRequested(void)
{
	// Tell our own application to quit.
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

// ===========================================================================
// MessageReceived
// ===========================================================================

void
MainWindow::MessageReceived(BMessage* message)
{
	// Figure out what kind of message was received.
	switch(message->what)
	{
		case MSG_REFRESH:
		{
			Refresh();
			CleanUp();
			break;
		}
	
		case MSG_CLEAN_UP:
		{
			CleanUp();
			break;
		}
		
		case B_ABOUT_REQUESTED:
		{
			be_app->PostMessage(message);
			break;
		}
		
		// Let the original BWindow handle the message.
		default:
		{
			BWindow::MessageReceived(message);
			break;
		}
	}
}

// ===========================================================================
// FrameMoved
// ===========================================================================

void
MainWindow::FrameMoved(BPoint screenPoint)
{
	// The little trick from FrameResized doesn't work from FrameMoved.
	SaveSettings();
}

// ===========================================================================
// FrameResized
// ===========================================================================

void
MainWindow::FrameResized(float width, float height)
{ 
	BPoint point;
	uint32 buttons;
	liveNodesView->GetMouse(&point,&buttons,false);
	if (!buttons)
	{
		SaveSettings();
	}
}

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ===========================================================================
// SaveSettings
// ===========================================================================

void
MainWindow::SaveSettings(void)
{
	BFile file(SETTINGS_FILE, B_CREATE_FILE | B_WRITE_ONLY);
	if (file.InitCheck() == B_OK)
	{
		Lock();
		if (file.Lock() == B_OK)
		{
			BRect rect = Frame();
			file.WriteAttr("MainWindowFrame",
				B_RECT_TYPE,
				0,
				&rect,
				sizeof(BRect));
      
			file.Sync();
			file.Unlock();
		}
		Unlock();
	}
}

// ===========================================================================
// Refresh
// ===========================================================================

void
MainWindow::Refresh(void)
{
	// Get rid of the old node views and connections.
	GetRidOfStuff();
	
	// Clear the information window and forget the node or connector
	// that was selected (because they don't exist anymore).
	infoWindow->SelectNothing();

	// -----------------------------------------------------------------------

	// Allocate new lists to store the node views and the connections.
	liveNodes = new BList(10);
	connections = new BList(10);
	
	// Because we don't know the number of live nodes beforehand, we'll use
	// a trick to get this number. We'll call BMediaRoster's GetLiveNodes()
	// twice; the first time to get the number of live nodes--which we can
	// use to allocate enough memory to hold the list--the second time to
	// actually get the list. Note that there could be a race condition here,
	// because nodes may have been added or removed inbetween the two calls
	// to GetLiveNodes().

	status_t result;
	live_node_info fakeLiveNode;
	int32 numLiveNodes = 1;
	result = mediaRoster->GetLiveNodes(&fakeLiveNode, &numLiveNodes);

	if (result != B_OK)
	{
		BBG_WRITE(("[MainWindow::Refresh] Can't get the list"
			" of live nodes. (%s)", strerror(result)));

		Fail("Sorry, but the Media Server won't give me the list of"
			" live nodes.");
	}
	else
	{
		BBG_WRITE(("[MainWindow::Refresh] There are probably %ld"
			" live nodes.", numLiveNodes));
	
		live_node_info* temp = (live_node_info*)
			malloc(numLiveNodes * sizeof(live_node_info));

		result = mediaRoster->GetLiveNodes(temp, &numLiveNodes);

		if (result != B_OK)
		{
			BBG_WRITE(("[MainWindow::GetAllLiveNodes] Can't get the list"
				" of live nodes. (%s)", strerror(result)));

			Fail("Sorry, but the Media Server won't give me the list of"
				" live nodes.");
		}
		else
		{
			// Successfully got the list of live nodes.
			BBG_WRITE(("[MainWindow::Refresh] Got %ld live"
				" nodes.", numLiveNodes));

			// ---------------------------------------------------------------

			BList* allConnectedInputs = new BList(10);
			BList* allConnectedOutputs = new BList(10);

			for (int32 t=0; t<numLiveNodes; ++t)
			{
				// Create this node's LiveNodeView object.
				LiveNodeView* view = new LiveNodeView(
					infoWindow,
					liveNodesView,
					temp[t]);
				
				// Add it to the list of node views.
				liveNodes->AddItem(view);

				// Get the node's input connectors, and add those who are
				// connected to the list of connected inputs.
				BList* inputConnectors = view->GetInputConnectors();
				if (inputConnectors != 0)
				{
					for (int32 t=0; t<inputConnectors->CountItems(); ++t)
					{
						ConnectorView* connectorView =
							(ConnectorView*) inputConnectors->ItemAt(t);
						if (connectorView->IsConnected())
						{
							allConnectedInputs->AddItem(connectorView);
						}
					}
				}

				// Get the node's output connectors, and add those who are
				// connected to the list of connected outputs.
				BList* outputConnectors = view->GetOutputConnectors();
				if (outputConnectors != 0)
				{
					for (int32 t=0; t<outputConnectors->CountItems(); ++t)
					{
						ConnectorView* connectorView =
							(ConnectorView*) outputConnectors->ItemAt(t);
						if (connectorView->IsConnected())
						{
							allConnectedOutputs->AddItem(connectorView);
						}
					}
				}
			}
			
			// ---------------------------------------------------------------
			
			// Loop through all the connected outputs and try
			// to find the inputs they are connected to.
			for (int32 outputIndex=0;
				outputIndex < allConnectedOutputs->CountItems();
				++outputIndex)
			{
				ConnectorView* outputConnector = (ConnectorView*)
					allConnectedOutputs->ItemAt(outputIndex);
					
				media_output* output = outputConnector->GetOutput();
						
				BBG_WRITE(("[MainWindow::Refresh] Output %ld is %s",
					outputIndex, output->name));

				for (int32 inputIndex=0;
					inputIndex < allConnectedInputs->CountItems();
					++inputIndex)
				{
					ConnectorView* inputConnector = (ConnectorView*)
						allConnectedInputs->ItemAt(inputIndex);
						
					media_input* input = inputConnector->GetInput();

					BBG_WRITE(("[MainWindow::Refresh] Trying to match to"
						" input %ld (%s)", inputIndex, input->name));
					
					// Source and destination id's are no good (since they
					// are local to nodes), but the actual source and
					// destination objects will do just fine.
					if ((input->source == output->source)
					&&  (input->destination == output->destination))
					{
						BBG_WRITE(("[MainWindow::Refresh] Found"
							" connection."));

						Connection* connection = new Connection(
							outputConnector,
							inputConnector);

						connections->AddItem(connection);

						// Done with this output.
						break;
					}
				}
			}

			// ---------------------------------------------------------------

			// Add the node views to the window.
			if (Lock())
			{
				if (liveNodes->CountItems() > 0)
				{
					for (int32 t=0; t<liveNodes->CountItems(); ++t)
					{
						LiveNodeView* view =
							(LiveNodeView*) liveNodes->ItemAt(t);
						liveNodesView->AddChild(view);
					}
				
					liveNodesView->SetConnections(connections);
				}
				
				liveNodesView->RedrawConnections();
				
				Unlock();
			}

			// ---------------------------------------------------------------
			
			// Delete the lists, but not their contents (which are still
			// being used elsewhere).
			delete allConnectedInputs;
			allConnectedInputs = 0;
			delete allConnectedOutputs;
			allConnectedOutputs = 0;
		}

		// We can safely free the list of live nodes here, because we've
		// passed the structures by value into the LiveNodeView constructors.
		free(temp);
	}

	// -----------------------------------------------------------------------

	// Allocate space to store the dormant node views.
	dormantNodes = new BList(10);

	dormant_node_info fakeDormantNode;
	int32 numDormantNodes = 1;
	result = mediaRoster->GetDormantNodes(&fakeDormantNode, &numDormantNodes);

	if (result != B_OK)
	{
		BBG_WRITE(("[MainWindow::Refresh] Can't get the list"
			" of dormant nodes. (%s)", strerror(result)));

		Fail("Sorry, but the Media Server won't give me the list of"
			" dormant nodes.");
	}
	else
	{
		BBG_WRITE(("[MainWindow::Refresh] There are probably %ld"
			" dormant nodes.", numDormantNodes));
	
		dormant_node_info* temp = (dormant_node_info*)
			malloc(numDormantNodes * sizeof(dormant_node_info));

		result = mediaRoster->GetDormantNodes(temp, &numDormantNodes);

		if (result != B_OK)
		{
			BBG_WRITE(("[MainWindow::Refresh] Can't get the list"
				" of dormant nodes. (%s)", strerror(result)));

			Fail("Sorry, but the Media Server won't give me the list of"
				" dormant nodes.");
		}
		else
		{
			// Successfully got the list of dormant nodes.
			BBG_WRITE(("[MainWindow::Refresh] Got %ld dormant"
				" nodes.", numDormantNodes));

			if (Lock())
			{
				for (int32 t=0; t<numDormantNodes; ++t)
				{
					// Create this node's DormantNodeView object.
					DormantNodeView* view = new DormantNodeView(
						infoWindow,
						temp[t]);
			
					// Add it to the list of node views.
					dormantNodes->AddItem(view);

					// And to the window.					
					dormantNodesView->AddChild(view);
				}
				Unlock();
			}
		}
		free(temp);
	}
	
	// -----------------------------------------------------------------------

	// Select a node.
	if (liveNodes->CountItems() > 0)
	{
		infoWindow->SelectNode((LiveNodeView*) liveNodes->ItemAt(0));
	}
	else if (dormantNodes->CountItems() > 0)
	{
		infoWindow->SelectNode((DormantNodeView*) dormantNodes->ItemAt(0));
	}
	
	// -----------------------------------------------------------------------
	
	return;
}

// ===========================================================================
// CleanUp
// ===========================================================================

void
MainWindow::CleanUp(void)
{
	/*!
	 * [TODO] MainWindow::CleanUp - Improve placement algorithm.
	 *
	 * Build a proper hierarchy instead of the brain-dead four columns
	 * approach I'm using right now. Is hint_point from live_node_info
	 * meant for this? (Surprisingly, hint_point arranges the live nodes
	 * exactly like I do, only with slighly smaller coordinates.)
	 */

	// Lock the window because we will be changing its views.
	if (Lock())
	{
		// This code places the nodes according to their hint_points.
		//for (int32 t=0; t<liveNodes->CountItems(); ++t)
		//{
		//	LiveNodeView* view = (LiveNodeView*) liveNodes->ItemAt(t);
		//	view->MoveTo(view->GetLiveNodeInfo().hint_point);
		//}
		//liveNodesView->Invalidate();
		//Unlock();
		//return;

		// -------------------------------------------------------------------
		
		// Column 1 stores the producers.
		// Column 2 stores the producers/consumers.
		// Column 3 stores the consumers.
		// Column 4 stores the rest.
				
		float columnWidth = liveNodesView->Bounds().Width()/4;

		float col1Center = columnWidth*0 + columnWidth/2;
		float col2Center = columnWidth*1 + columnWidth/2;
		float col3Center = columnWidth*2 + columnWidth/2;
		float col4Center = columnWidth*3 + columnWidth/2;
			
		float col1Y = 8;
		float col2Y = 8;
		float col3Y = 8;
		float col4Y = 8;
			
		for (int32 t=0; t<liveNodes->CountItems(); ++t)
		{
			LiveNodeView* view = (LiveNodeView*) liveNodes->ItemAt(t);
			int32 kind = view->GetNode().kind;
			
			if ((kind & B_BUFFER_PRODUCER)
			&& !(kind & B_BUFFER_CONSUMER))
			{
				// We need to truncate (or round off) the x-coordinate to
				// prevent rounding-errors when moving the connectors.
				int32 x = (int32)(col1Center - view->Bounds().Width()/2.0);

				view->MoveTo(BPoint(x, col1Y));
				col1Y += view->Bounds().Height()+8;
			}
			else if ((kind & B_BUFFER_PRODUCER)
				 &&  (kind & B_BUFFER_CONSUMER))
			{
				int32 x = (int32)(col2Center - view->Bounds().Width()/2.0);
				view->MoveTo(BPoint(x, col2Y));
				col2Y += view->Bounds().Height()+8;
			}
			else if (!(kind & B_BUFFER_PRODUCER)
				 &&   (kind & B_BUFFER_CONSUMER))
			{
				int32 x = (int32)(col3Center - view->Bounds().Width()/2.0);
				view->MoveTo(BPoint(x, col3Y));
				col3Y += view->Bounds().Height()+8;
			}
			else
			{
				int32 x = (int32)(col4Center - view->Bounds().Width()/2.0);
				view->MoveTo(BPoint(x, col4Y));
				col4Y += view->Bounds().Height()+8;
			}
		}

		// Redraw the live nodes view.
		liveNodesView->Invalidate();

		// -------------------------------------------------------------------

		float y = 10;
		float centerX = dormantNodesView->Bounds().Width()/2;
		
		for (int32 t=0; t<dormantNodes->CountItems(); ++t)
		{
			DormantNodeView* view =
				(DormantNodeView*) dormantNodes->ItemAt(t);

			view->MoveTo(BPoint(centerX - view->Bounds().Width()/2, y));
			
			y += view->Bounds().Height() + 10;
		}

		// Redraw the dormant nodes view.
		dormantNodesView->Invalidate();

		// -------------------------------------------------------------------
						
		// We're done with the window now, so we should unlock it.
		Unlock();
	}
}

// ===========================================================================
// GetRidOfStuff
// ===========================================================================

void
MainWindow::GetRidOfStuff(void)
{
	// Get rid of the live nodes.
	if (liveNodes != 0)
	{
		if (Lock())
		{
			int32 countItems = liveNodes->CountItems();
			for (int32 t=0; t<countItems; ++t)
			{
				BView* view = (BView*) liveNodes->ItemAt(t);
				bool result = view->RemoveSelf();
				delete view;
		    }
			delete liveNodes;
			liveNodes = 0;

			Unlock();
		}
	}

	// Get rid of the connections.
	if (connections != 0)
	{
		int32 countItems = connections->CountItems();
		for (int32 t=0; t<countItems; ++t)
		{
			Connection* connection = (Connection*) connections->ItemAt(t);
			delete connection;
		}
		delete connections;
		connections = 0;
		liveNodesView->SetConnections(0);
	}

	// Get rid of the dormant nodes.
	if (dormantNodes != 0)
	{
		if (Lock())
		{
			int32 countItems = dormantNodes->CountItems();
			for (int32 t=0; t<countItems; ++t)
			{
				BView* view = (BView*) dormantNodes->ItemAt(t);
				bool result = view->RemoveSelf();
				delete view;
		    }
			delete dormantNodes;
			dormantNodes = 0;

			Unlock();
		}
	}
}

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
